/*
 * The DSP56001 Device Driver, saviour of the Free World(tm)
 *
 * Authors: Fredrik Noring   <noring@lysator.liu.se>
 *          lars brinkhoff   <f93labr@dd.chalmers.se>
 *          Tomas Berndtsson <tobe@lysator.liu.se>
 *
 * Written May 1996
 *
 * It's now June 1996, and it's still being written, but now with
 * 14 megs of accessible memory. What a relief.
 *
 * September 1996. I found the sources stuffed away in my fridge.
 *
 * December 1996. I, Tomas, brushed off the mould growing on the
 * sources. First attempts of implementing DMA->DSP transfer.
 * The source never looked the same after I've put my hands in it.
 *
 * It's January 1997 and I start to program on this device driver
 * again. The driver has alot more features now, and the structure
 * is getting better. All codec code is removed (and found in the
 * new device driver /dev/matrix). Moving closer to a 1.0 release.
 *
 * BUGS:
 *  Hmm... there must be something here :)
 *
 * Copyright (C) 1996,1997 Fredrik Noring, lars brinkhoff & Tomas Berndtsson
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive
 * for more details.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/malloc.h>	/* for kmalloc() and kfree() */
#include <linux/sched.h>	/* for struct wait_queue etc */
#include <linux/major.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/delay.h>	/* guess what */
#include <linux/fs.h>
#include <linux/mm.h>

#include <asm/segment.h>
#include <asm/atarihw.h>
#include <asm/traps.h>

#ifdef 0
#include <asm/dsp56k.h>
#endif
#include "dsp56k.h"

/* should be in include/linux/major.h */
#define DSP56K_MAJOR            55

/* minor devices */
#define DSP56K_DEV_56001        0    /* The only device so far */



#define DEBUG_LEVEL 1   /* Will be removed! */
#if DEBUG_LEVEL > 0
#define DEBUG(n, x, y...) \
({ \
	   if(n <= DEBUG_LEVEL) \
		   printk("%s: %s(): "x"\n", __FILE__, __FUNCTION__ , ## y); \
})
#else
#define DEBUG(n, x, y...)
#endif

#define FATAL(x, y...)	DEBUG(1, x , ## y)
#define ERROR(x, y...)	DEBUG(2, x , ## y)
#define INFO(x, y...)	DEBUG(3, x , ## y)
#define BLAH(x, y...)	DEBUG(4, x , ## y)




#define TIMEOUT    10   /* Host port timeout in number of tries */
#define MAXIO    2048   /* Maximum number of words before sleep */
#define DSP56K_MAX_BINARY_LENGTH (3*64*1024)


#define DSP56K_TX_INT_ON	dsp56k_host_interface.icr |=  DSP56K_ICR_TREQ
#define DSP56K_RX_INT_ON	dsp56k_host_interface.icr |=  DSP56K_ICR_RREQ
#define DSP56K_TX_INT_OFF	dsp56k_host_interface.icr &= ~DSP56K_ICR_TREQ
#define DSP56K_RX_INT_OFF	dsp56k_host_interface.icr &= ~DSP56K_ICR_RREQ

#define DSP56K_TRANSMIT		(dsp56k_host_interface.isr & DSP56K_ISR_TXDE)
#define DSP56K_RECEIVE		(dsp56k_host_interface.isr & DSP56K_ISR_RXDF)

#define max(a,b) ((a) > (b) ? (a) : (b))
#define min(a,b) ((a) < (b) ? (a) : (b))

#define wait_some(n) \
{ \
	current->state = TASK_INTERRUPTIBLE; \
	current->timeout = jiffies + n; \
	schedule(); \
}

#define handshake(count, maxio, timeout, ENABLE, f) \
{ \
	long i, t, m; \
	while (count > 0) { \
		m = min(count, maxio); \
		for (i = 0; i < m; i++) { \
			for (t = 0; t < timeout && !ENABLE; t++) \
				wait_some(2); \
			if(!ENABLE) \
				return -EIO; \
			f; \
		} \
		count -= m; \
		if (m == maxio) wait_some(2); \
	} \
}

#define tx_wait(n, message) \
{ \
	int t; \
	for(t = 0; t < n && !DSP56K_TRANSMIT; t++) \
		wait_some(1); \
	if(!DSP56K_TRANSMIT && *message) { \
		ERROR(message); \
		return -EIO; \
	} \
}

#define rx_wait(n, message) \
{ \
	int t; \
	for(t = 0; t < n && !DSP56K_RECEIVE; t++) \
		wait_some(1); \
	if(!DSP56K_RECEIVE && *message) { \
		ERROR(message); \
		return -EIO; \
	} \
}

/* DSP56001 bootstrap code */
static char bootstrap[] = {
0x0c, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
0x00, 0x00, 0x60, 0xf4, 0x00, 0x00, 0x00, 0x4f, 0x61, 0xf4,
0x00, 0x00, 0x7e, 0xa9, 0x06, 0x2e, 0x80, 0x00, 0x00, 0x47,
0x07, 0xd8, 0x84, 0x07, 0x59, 0x84, 0x08, 0xf4, 0xa8, 0x00,
0x00, 0x04, 0x08, 0xf4, 0xbf, 0x00, 0x0c, 0x00, 0x00, 0xfe,
0xb8, 0x0a, 0xf0, 0x80, 0x00, 0x7e, 0xa9, 0x08, 0xf4, 0xa0,
0x00, 0x00, 0x01, 0x08, 0xf4, 0xbe, 0x00, 0x00, 0x00, 0x0a,
0xa9, 0x80, 0x00, 0x7e, 0xad, 0x08, 0x4e, 0x2b, 0x44, 0xf4,
0x00, 0x00, 0x00, 0x03, 0x44, 0xf4, 0x45, 0x00, 0x00, 0x01,
0x0e, 0xa0, 0x00, 0x0a, 0xa9, 0x80, 0x00, 0x7e, 0xb5, 0x08,
0x50, 0x2b, 0x0a, 0xa9, 0x80, 0x00, 0x7e, 0xb8, 0x08, 0x46,
0x2b, 0x44, 0xf4, 0x45, 0x00, 0x00, 0x02, 0x0a, 0xf0, 0xaa,
0x00, 0x7e, 0xc9, 0x20, 0x00, 0x45, 0x0a, 0xf0, 0xaa, 0x00,
0x7e, 0xd0, 0x06, 0xc6, 0x00, 0x00, 0x7e, 0xc6, 0x0a, 0xa9,
0x80, 0x00, 0x7e, 0xc4, 0x08, 0x58, 0x6b, 0x0a, 0xf0, 0x80,
0x00, 0x7e, 0xad, 0x06, 0xc6, 0x00, 0x00, 0x7e, 0xcd, 0x0a,
0xa9, 0x80, 0x00, 0x7e, 0xcb, 0x08, 0x58, 0xab, 0x0a, 0xf0,
0x80, 0x00, 0x7e, 0xad, 0x06, 0xc6, 0x00, 0x00, 0x7e, 0xd4,
0x0a, 0xa9, 0x80, 0x00, 0x7e, 0xd2, 0x08, 0x58, 0xeb, 0x0a,
0xf0, 0x80, 0x00, 0x7e, 0xad};
static int sizeof_bootstrap = 375;


static struct dsp56k_device {
  int in_use;
  long maxio, timeout;
  int tx_wsize, rx_wsize;
} dsp56k;

static int dsp56k_reset(void)
{
  u_char status;
	
  BLAH("resetting dsp...");
	
  /* Power down the DSP */
  sound_ym.rd_data_reg_sel = 14;
  status = sound_ym.rd_data_reg_sel & 0xef;
  sound_ym.wd_data = status;
  sound_ym.wd_data = status | 0x10;
  
  udelay(10);
  
  /* Power up the DSP */
  sound_ym.rd_data_reg_sel = 14;
  sound_ym.wd_data = sound_ym.rd_data_reg_sel & 0xef;

  INFO("dsp reset");
  return 0;
}

static int dsp56k_upload(u_char *bin, int len)
{
  int i;
  u_char *p;
	
  BLAH("bootstrapping the dsp...");
	
  dsp56k_reset();
  
  p = bootstrap;
  for (i = 0; i < sizeof_bootstrap/3; i++) {
    /* tx_wait(10, "dsp not recieving bootstrap code"); */
    dsp56k_host_interface.data.b[1] = *p++;
    dsp56k_host_interface.data.b[2] = *p++;
    dsp56k_host_interface.data.b[3] = *p++;
  }
  for (; i < 512; i++) {
    /* tx_wait(10, "dsp not recieving bootstrap padding"); */
    dsp56k_host_interface.data.b[1] = 0;
    dsp56k_host_interface.data.b[2] = 0;
    dsp56k_host_interface.data.b[3] = 0;
  }
  BLAH("bootstrap code sent");
  
  for (i = 0; i < len; i++) {
    tx_wait(10, "dsp not recieving user code");
    dsp56k_host_interface.data.b[1] = get_user(bin++);
    dsp56k_host_interface.data.b[2] = get_user(bin++);
    dsp56k_host_interface.data.b[3] = get_user(bin++);
  }
  BLAH("user code sent");

  tx_wait(10, "dsp not recieving start command");
  dsp56k_host_interface.data.l = 3;    /* Magic execute */

  BLAH("dsp bootstrap done.");
  return 0;
}

static int dsp56k_read(struct inode *inode, struct file *file,
			char *buf, int count)
{
  int dev = MINOR(inode->i_rdev) & 0x0f;

  switch(dev)
    {
    case DSP56K_DEV_56001:
      {

	long n;

    /* Don't do anything if nothing is to be done */
	if (!count) return 0;

	n = 0;
	switch (dsp56k.rx_wsize) {
	case 1:  /* 8 bit */
	  {
	    handshake(count, dsp56k.maxio, dsp56k.timeout, DSP56K_RECEIVE,
		      put_user(dsp56k_host_interface.data.b[3], buf+n++));
	    return n;
	  }
	case 2:  /* 16 bit */
	  {
	    short *data;

	    count /= 2;
	    data = (short*) buf;
	    handshake(count, dsp56k.maxio, dsp56k.timeout, DSP56K_RECEIVE,
		      put_user(dsp56k_host_interface.data.w[1], data+n++));
	    return 2*n;
	  }
	case 3:  /* 24 bit */
	  {
	    count /= 3;
	    handshake(count, dsp56k.maxio, dsp56k.timeout, DSP56K_RECEIVE,
		      put_user(dsp56k_host_interface.data.b[1], buf+n++);
		      put_user(dsp56k_host_interface.data.b[2], buf+n++);
		      put_user(dsp56k_host_interface.data.b[3], buf+n++));
	    return 3*n;
	  }
	case 4:  /* 32 bit */
	  {
	    long *data;

	    count /= 4;
	    data = (long*) buf;
	    handshake(count, dsp56k.maxio, dsp56k.timeout, DSP56K_RECEIVE,
		      put_user(dsp56k_host_interface.data.l, data+n++));
	    return 4*n;
	  }
	}
	return -EFAULT;
      }

    default:
      ERROR("Unknown minor device");
      return -ENXIO;
    }
}

static int dsp56k_write(struct inode *inode, struct file *file,
			 const char *buf, int count)
{
  int dev = MINOR(inode->i_rdev) & 0x0f;

  switch(dev)
    {
    case DSP56K_DEV_56001:
      {
	long n;

    /* Don't do anything if nothing is to be done */
	if (!count) return 0;

	n = 0;
	switch (dsp56k.tx_wsize) {
	case 1:  /* 8 bit */
	  {
	    handshake(count, dsp56k.maxio, dsp56k.timeout, DSP56K_TRANSMIT,
		      {dsp56k_host_interface.data.b[3]=get_user(buf+n++);});
	    return n;
	  }
	case 2:  /* 16 bit */
	  {
	    short *data;

	    count /= 2;
	    data = (short*) buf;
	    handshake(count, dsp56k.maxio, dsp56k.timeout, DSP56K_TRANSMIT,
		      {dsp56k_host_interface.data.w[1]=get_user(data+n++);});
	    return 2*n;
	  }
	case 3:  /* 24 bit */
	  {
	    count /= 3;
	    handshake(count, dsp56k.maxio, dsp56k.timeout, DSP56K_TRANSMIT,
		      {dsp56k_host_interface.data.b[1]=get_user(buf+n++);
		       dsp56k_host_interface.data.b[2]=get_user(buf+n++);
		       dsp56k_host_interface.data.b[3]=get_user(buf+n++);});
	    return 3*n;
	  }
	case 4:  /* 32 bit */
	  {
	    long *data;

	    count /= 4;
	    data = (long*) buf;
	    handshake(count, dsp56k.maxio, dsp56k.timeout, DSP56K_TRANSMIT,
		      {dsp56k_host_interface.data.l=get_user(data+n++);});
	    return 4*n;
	  }
	}

	return -EFAULT;
      }
    default:
      ERROR("Unknown minor device");
      return -ENXIO;
    }
}

static int dsp56k_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
  int dev = MINOR(inode->i_rdev) & 0x0f;

  switch(dev)
    {
    case DSP56K_DEV_56001:

      switch(cmd) {
      case DSP56K_UPLOAD:
	{
	  char *bin;
	  int r, len;
	  struct dsp56k_upload *binary = (struct dsp56k_upload *) arg;
    
	  INFO("DSP56K_UPLOAD");
		
	  len = get_user(&binary->len);
	  bin = get_user(&binary->bin);
	  BLAH("len = %d, bin = %p", len, bin);
		
	  if (len == 0) {
	    ERROR("len == 0, nothing to upload?");
	    return -EINVAL;      /* nothing to upload?!? */
	  }
	  if (len > DSP56K_MAX_BINARY_LENGTH) {
	    ERROR("user code too big");
	    return -EINVAL;
	  }
    
	  r = dsp56k_upload(bin, len);
	  if (r < 0) {
	    ERROR("dsp56k_upload failed");
	    return r;
	  }
    
	  break;
	}
      case DSP56K_SET_TX_WSIZE:
	if (arg > 4 || arg < 1)
	  return -EINVAL;
	dsp56k.tx_wsize = (int) arg;
	break;
      case DSP56K_SET_RX_WSIZE:
	if (arg > 4 || arg < 1)
	  return -EINVAL;
	dsp56k.rx_wsize = (int) arg;
	break;
      case DSP56K_HOST_FLAGS:
	{
	  int dir, out, status;
	  struct dsp56k_host_flags *hf = (struct dsp56k_host_flags*) arg;
    
	  INFO("DSP56K_HF");
		
	  dir = get_user(&hf->dir);
	  out = get_user(&hf->out);

	  if ((dir & 0x1) && (out & 0x1))
	    dsp56k_host_interface.icr |= DSP56K_ICR_HF0;
	  else if (dir & 0x1)
	    dsp56k_host_interface.icr &= ~DSP56K_ICR_HF0;
	  if ((dir & 0x2) && (out & 0x2))
	    dsp56k_host_interface.icr |= DSP56K_ICR_HF1;
	  else if (dir & 0x2)
	    dsp56k_host_interface.icr &= ~DSP56K_ICR_HF1;

	  status = 0;
	  if (dsp56k_host_interface.icr & DSP56K_ICR_HF0) status |= 0x1;
	  if (dsp56k_host_interface.icr & DSP56K_ICR_HF1) status |= 0x2;
	  if (dsp56k_host_interface.isr & DSP56K_ISR_HF2) status |= 0x4;
	  if (dsp56k_host_interface.isr & DSP56K_ISR_HF3) status |= 0x8;

	  put_user(status, &hf->status);
	  break;
	}
      case DSP56K_HOST_CMD:
	if (arg > 31 || arg < 0)
	  return -EINVAL;
	dsp56k_host_interface.cvr = (u_char)((arg & DSP56K_CVR_HV_MASK) |
					     DSP56K_CVR_HC);
	break;
      default:
	ERROR("Unknown ioctl");
	return -EINVAL;
      }
      return 0;

    default:
      ERROR("Unknown minor device");
      return -ENXIO;
    }
}

/* should be dsp56k_poll, if needed at all */
#ifdef 0
static int dsp56k_select(struct inode *inode, struct file *file, int sel_type,
			 select_table *wait)
{
  int dev = MINOR(inode->i_rdev) & 0x0f;

  switch(dev)
    {
    case DSP56K_DEV_56001:

      switch(sel_type) {
      case SEL_IN:		/* read */
	return 1;
      case SEL_OUT:		/* write */
	return 1;
      default:
	ERROR("unknown sel_type");
	return 1;
      }

    default:
      ERROR("Unknown minor device");
      return -ENXIO;
    }
}
#endif

static int dsp56k_open(struct inode *inode, struct file *file)
{
  int dev = MINOR(inode->i_rdev) & 0x0f;

  switch(dev)
    {
    case DSP56K_DEV_56001:
      INFO("called");

      if (dsp56k.in_use) {
	BLAH("dsp56k busy");
	return -EBUSY;
      }

      BLAH("dsp56k free");

      dsp56k.in_use = 1;
      dsp56k.timeout = TIMEOUT;
      dsp56k.maxio = MAXIO;
      dsp56k.rx_wsize = dsp56k.tx_wsize = 4; 

      DSP56K_TX_INT_OFF;
      DSP56K_RX_INT_OFF;

      /* Zero host flags */
      dsp56k_host_interface.icr &= ~DSP56K_ICR_HF0;
      dsp56k_host_interface.icr &= ~DSP56K_ICR_HF1;

      break;

    default:
      ERROR("Unknown minor device");
      return -ENXIO;
    }

#ifdef MODULE
      MOD_INC_USE_COUNT;
#endif /* MODULE */

      return 0;
}

static void dsp56k_release(struct inode *inode, struct file *file)
{
  int dev = MINOR(inode->i_rdev) & 0x0f;

  switch(dev)
    {
    case DSP56K_DEV_56001:
      INFO("called");

      if (!dsp56k.in_use)
	ERROR("device closed without being used?!?");

      dsp56k.in_use = 0;

      break;
    default:
      ERROR("Unknown minor device");
      return;
    }

#ifdef MODULE
      MOD_DEC_USE_COUNT;
#endif /* MODULE */
}

static struct file_operations dsp56k_fops = {
  NULL,    /* no special dsp56k_lseek */
  dsp56k_read,
  dsp56k_write,
  NULL,    /* no special dsp56k_readdir */
  NULL,    /* dsp56k_poll? */
  dsp56k_ioctl,
  NULL,    /* no special dsp56k_mmap */
  dsp56k_open,
  dsp56k_release,
  NULL,    /* no special dsp56k_fsync */
  NULL,    /* no special dsp56k_fasync */
  NULL,    /* no special dsp56k_check_media_change */
  NULL     /* no special dsp56k_revalidate */
};


/****** Config & setup ******/

static int setup_error = 0;

void dsp56k_setup(char *str, int *ints)
{
#ifndef MODULE
  int r;

  r = register_chrdev(DSP56K_MAJOR, "dsp56k", &dsp56k_fops);
  if(r) {
    printk("DSP56k driver: Unable to register driver\n");
    return;
  }
#endif /* MODULE */  

  dsp56k.in_use = 0;

  printk("DSP56k driver installed\n");
}

#ifdef MODULE
int init_module(void)
{
  int r;

  INFO("called");

  setup_error = 0;
  dsp56k_setup(NULL, NULL);
  if(setup_error)
    return -EPERM;

  r = register_chrdev(DSP56K_MAJOR, "dsp56k", &dsp56k_fops);
  if(r) {
    printk("DSP56k driver: Unable to register driver\n");
    return r;
  }
	
  return 0;
}

void cleanup_module(void)
{
  INFO("called");

  unregister_chrdev(DSP56K_MAJOR, "dsp56k");
}
#endif /* MODULE */
